#include <osg/NodeCallback>
#include <osg/Timer>
#include <osgCal/Model>
#include <osgCal/SubMesh>

using namespace osgCal;

/** @class CalUpdateCallback
 * @brief Class that updates the models to generate the new deformed vertices
 */
class CalUpdateCallback: public osg::NodeCallback {
public:
    CalUpdateCallback(CalModel *_model): model(_model) {}
    virtual void operator()(osg::Node* node, osg::NodeVisitor* nv)
    {
	osg::Timer_t current = timer.tick();
        double elapsed = timer.delta_s(previous, current);
        model->update(elapsed);
        previous=current;
	traverse(node,nv);
    }
private:
    CalModel *model;
    osg::Timer timer;
    osg::Timer_t previous;
};

Model::Model(){}

Model::Model(CoreModel *core) {
    create(core);
}

void Model::create(CoreModel *core) {
    if (!model.create(core->get())) {
	std::cerr << "Model::create: ERROR: Can't create a model. Most probable reason is that the core doesn't have any data." << std::endl;
	CalError::printLastError();
        return;
    }
    setUpdateCallback(new CalUpdateCallback(&model));
    // attach all meshes to the model
    for(int meshId = 0; meshId < core->get()->getCoreMeshCount(); meshId++)
    {
	model.attachMesh(meshId);
    }
    model.setMaterialSet(0);

    CalRenderer *pCalRenderer = model.getRenderer();
    if(pCalRenderer->beginRendering()) {
	int meshCount = pCalRenderer->getMeshCount();
	for(int meshId = 0; meshId < meshCount; meshId++) {
	    int submeshCount = pCalRenderer->getSubmeshCount(meshId);
	    for(int submeshId = 0; submeshId < submeshCount; submeshId++) {
		SubMesh *submesh = new SubMesh(&model, meshId, submeshId);
		addDrawable(submesh);
	    }
	}
	pCalRenderer->endRendering();
    }

    /// Force generation of first mesh
    model.update(0);
    this->coreModel=core;
}

void Model::action(unsigned id, float delay_in, float delay_out) {
    if (model.getMixer() != 0)
	model.getMixer()->executeAction(id, delay_in, delay_out);
    else std::cerr << "Model::action: ERROR: The model doesn't have a mixer" << std::endl;
}

void Model::loop(unsigned id, float weight, float delay) {
    if (model.getMixer() != 0)
	model.getMixer()->blendCycle(id, weight, delay);
    else std::cerr << "Model::loop: ERROR: The model doesn't have a mixer" << std::endl;
}

void Model::stop(unsigned id, float delay) {
    if (model.getMixer() != 0)
	model.getMixer()->clearCycle(id, delay);
    else std::cerr << "Model::stop: ERROR: The model doesn't have a mixer" << std::endl;
}

osg::Object* Model::clone(const osg::CopyOp&) const {
    return new Model();
}
