#include <osg/Material>
#include <osg/Texture2D>
#include <osgCal/SubMesh>

using namespace osgCal;

SubMesh::SubMesh() {
    std::cout << "SubMesh::SubMesh(): You should never call this constructor!" << std::endl;
    exit(1);
}

SubMesh::SubMesh(CalModel *model, unsigned mesh, unsigned submesh) {
    this->model=model;
    this->mesh=mesh;
    this->submesh=submesh;

    setUseDisplayList(false);
    setUpMaterial();
}

void SubMesh::setUpMaterial() {
    CalRenderer *pCalRenderer;
    pCalRenderer = model->getRenderer();
    osg::StateSet *set = this->getOrCreateStateSet();
    if(pCalRenderer->beginRendering()) {
	// select mesh and submesh for further data access
	if(pCalRenderer->selectMeshSubmesh(mesh, submesh)) {
	    osg::Material *material = new osg::Material();
	    set->setAttributeAndModes(material, osg::StateAttribute::ON);
	    unsigned char meshColor[4];
	    osg::Vec4 materialColor;

	    // set the material ambient color
	    pCalRenderer->getAmbientColor(&meshColor[0]);
	    materialColor[0] = meshColor[0] / 255.0f;  materialColor[1] = meshColor[1] / 255.0f; materialColor[2] = meshColor[2] / 255.0f; materialColor[3] = meshColor[3] / 255.0f;
            if (materialColor[3] == 0) materialColor[3]=1.0f;
	    material->setAmbient(osg::Material::FRONT, materialColor);

	    // set the material diffuse color
	    pCalRenderer->getDiffuseColor(&meshColor[0]);
	    materialColor[0] = meshColor[0] / 255.0f;  materialColor[1] = meshColor[1] / 255.0f; materialColor[2] = meshColor[2] / 255.0f; materialColor[3] = meshColor[3] / 255.0f;
            if (materialColor[3] == 0) materialColor[3]=1.0f;
	    material->setDiffuse(osg::Material::FRONT, materialColor);

	    // set the material specular color
	    pCalRenderer->getSpecularColor(&meshColor[0]);
	    materialColor[0] = meshColor[0] / 255.0f;  materialColor[1] = meshColor[1] / 255.0f; materialColor[2] = meshColor[2] / 255.0f; materialColor[3] = meshColor[3] / 255.0f;
            if (materialColor[3] == 0) materialColor[3]=1.0f;
	    material->setSpecular(osg::Material::FRONT, materialColor);

	    // set the material shininess factor
	    float shininess;
	    shininess = pCalRenderer->getShininess();
	    material->setShininess(osg::Material::FRONT, shininess);

	    if (pCalRenderer->getMapCount() > 0)
	    {
		osg::Texture2D *texture = (osg::Texture2D*)pCalRenderer->getMapUserData(0);
		if (texture != 0) {
		    set->setTextureAttributeAndModes(0, texture, osg::StateAttribute::ON);
		}
	    }
	}
	pCalRenderer->endRendering();
    }
}

void SubMesh::drawImplementation(osg::State& state) const {
    // get the renderer of the model
    CalRenderer *pCalRenderer;
    pCalRenderer = model->getRenderer();

    // begin the rendering loop
    if(pCalRenderer->beginRendering())
    {
	// we will use vertex arrays, so enable them
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	// select mesh and submesh for further data access
	if(pCalRenderer->selectMeshSubmesh(mesh, submesh))
	{
	    // get the transformed vertices of the submesh
	    int vertexCount = pCalRenderer->getVertexCount();
	    float *meshVertices = new float[vertexCount*3];
	    float *meshNormals = new float[vertexCount*3];

	    vertexCount = pCalRenderer->getVertices(meshVertices);

	    // get the transformed normals of the submesh
	    pCalRenderer->getNormals(meshNormals);
	    // get the texture coordinates of the submesh
	    //static float meshTextureCoordinates[30000][2];
            float *meshTextureCoordinates = new float[vertexCount*2];
	    unsigned tcount = pCalRenderer->getTextureCoordinates(0, meshTextureCoordinates);

	    // get the faces of the submesh
            CalIndex *meshFaces = new CalIndex[pCalRenderer->getFaceCount()*3];
            int faceCount;
	    faceCount = pCalRenderer->getFaces(meshFaces);

	    // set the vertex and normal buffers
	    glVertexPointer(3, GL_FLOAT, 0, meshVertices);
	    glNormalPointer(GL_FLOAT, 0, meshNormals);
	    // set the texture coordinate buffer and state if necessary
	    if((pCalRenderer->getMapCount() > 0) && (tcount > 0))
	    {
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnable(GL_COLOR_MATERIAL);

		// set the texture coordinate buffer
		glTexCoordPointer(2, GL_FLOAT, 0, meshTextureCoordinates);
		glColor3f(1.0f, 1.0f, 1.0f);
	    }

	    // draw the submesh

	    if(sizeof(CalIndex)==2)
		glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_SHORT, meshFaces);
	    else
		glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_INT, meshFaces);

	    // disable the texture coordinate state if necessary
	    if((pCalRenderer->getMapCount() > 0) && (tcount > 0))
	    {
		glDisable(GL_COLOR_MATERIAL);
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	    }
	    delete [] meshVertices;
	    delete [] meshNormals;
            delete [] meshFaces;
            delete [] meshTextureCoordinates;
	}

	// clear vertex array state
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);

	// end the rendering
	pCalRenderer->endRendering();
    }
}

osg::Object* SubMesh::clone(const osg::CopyOp&) const {
    return new SubMesh(model, mesh, submesh);
}

bool SubMesh::computeBound() const {
    _bbox.init();
    // We have to traverse all vertices to racalculate bounds
    CalRenderer *pCalRenderer;
    pCalRenderer = model->getRenderer();
    if(pCalRenderer->beginRendering()) {
	if(pCalRenderer->selectMeshSubmesh(mesh, submesh)) {
	    // get the transformed vertices of the submesh
	    int vertexCount = pCalRenderer->getVertexCount();
	    float *meshVertices = new float[vertexCount*3];
	    vertexCount = pCalRenderer->getVertices(meshVertices);
	    for (int vertex=0;vertex<vertexCount;vertex++) {
		float* v=meshVertices+vertex*3;
		_bbox.expandBy(v[0],v[1],v[2]);
	    }
	    delete []meshVertices;
	}
	pCalRenderer->endRendering();
    }
    _bbox_computed=false; // Force calculation on each frame
    return true;
}
